#include "Audio.h"
#include "NUtilDefs.h"



IDirectSound8* Audio::mDevice = NULL;

void Audio::Init(HWND hWnd)
{
	DirectSoundCreate8(NULL, &mDevice, NULL);
	mDevice->SetCooperativeLevel(hWnd, DSSCL_PRIORITY);
}

void Audio::Deinit()
{
	SafeRelease(mDevice);
}



void NWaveFileHeader::LoadWAVHeaderFromFile(FILE* file)
{

#define EZRead(x) fread(&x, sizeof(x), 1, file);

	ChunkID;//          Contains the letters "RIFF" in ASCII form
	//(0x52494646 big-endian form).
	EZRead(ChunkID);

	assert(ChunkID == 'FFIR');

	ChunkSize;//        36 + SubChunk2Size, or more precisely:
	//4 + (8 + SubChunk1Size) + (8 + SubChunk2Size)
	//	This is the size of the rest of the chunk
	//	following this number.  This is the size of the
	//	entire file in bytes minus 8 bytes for the
	//	two fields not included in this count:
	//ChunkID and ChunkSize.
	EZRead(ChunkSize);

	Format;//           Contains the letters "WAVE"
	//	(0x57415645 big-endian form).

	//	The "WAVE" format consists of two subchunks: "fmt " and "data":
	//The "fmt " subchunk describes the sound data's format:
	EZRead(Format);

	assert(Format == 'EVAW');

	Subchunk1ID;/*      Contains the letters "fmt "
				(0x666d7420 big-endian form).*/
	EZRead(Subchunk1ID);

	// Other chunks may exist here. Skip them.
	while (Subchunk1ID != ' tmf')
	{
		EZRead(Subchunk1Size);
		fseek(file, Subchunk1Size, SEEK_CUR);
		EZRead(Subchunk1ID);
	}

	assert(Subchunk1ID == ' tmf');

	Subchunk1Size;/*    16 for PCM.  This is the size of the
				  rest of the Subchunk which follows this number.*/
	EZRead(Subchunk1Size);

	AudioFormat;/*      PCM = 1 (i.e. Linear quantization)
				Values other than 1 indicate some
				form of compression.*/
	EZRead(AudioFormat);

	NumChannels;//      Mono = 1, Stereo = 2, etc.
	EZRead(NumChannels);

	SampleRate;//       8000, 44100, etc.
	EZRead(SampleRate);

	ByteRate;//         == SampleRate * NumChannels * BitsPerSample/8
	EZRead(ByteRate);

	BlockAlign;//       == NumChannels * BitsPerSample/8
	//The number of bytes for one sample including
	//all channels. I wonder what happens when
	//this number isn't an integer?
	EZRead(BlockAlign);

	BitsPerSample;//    8 bits = 8, 16 bits = 16, etc.
	//2   ExtraParamSize   if PCM, then doesn't exist
	//X   ExtraParams      space for extra parameters

	//The "data" subchunk contains the size of the data and the actual sound:
	EZRead(BitsPerSample);

	Subchunk2ID;/*     Contains the letters "data"
				(0x64617461 big-endian form).*/
	EZRead(Subchunk2ID);

	// Other chunks may exist here. Skip them.
	while (Subchunk2ID != 'atad')
	{
		EZRead(Subchunk2Size);
		fseek(file, Subchunk2Size, SEEK_CUR);
		EZRead(Subchunk2ID);
	}

	assert(Subchunk2ID == 'atad');

	Subchunk2Size;/*    == NumSamples * NumChannels * BitsPerSample/8
				  This is the number of bytes in the data.
				  You can also think of this as the size
				  of the read of the subchunk following this
				  number.*/
	//44        *   Data             The actual sound data.
	EZRead(Subchunk2Size);

#undef EZRead

}



Sound::Sound()
{
	memset(&Header, 0, sizeof(Header));
	memset(&Format, 0, sizeof(Format));
	memset(&BufferDesc, 0, sizeof(BufferDesc));
	Buffer = NULL;
	memset(&BufferCaps, 0, sizeof(BufferCaps));

	Format.wFormatTag = WAVE_FORMAT_PCM;
}

Sound::~Sound()
{
}

void Sound::Load(const char* Filename)
{
	FILE* InFile = NULL;
	fopen_s(&InFile, Filename, "rb");

	Header.LoadWAVHeaderFromFile(InFile);

	Format.nChannels = Header.NumChannels;
	Format.wBitsPerSample = Header.BitsPerSample;
	Format.nSamplesPerSec = Header.SampleRate;

	Format.nBlockAlign = (Format.nChannels * Format.wBitsPerSample) / 8;
	Format.nAvgBytesPerSec = Format.nSamplesPerSec * Format.nBlockAlign;

	BufferDesc.dwSize = sizeof(BufferDesc);
	BufferDesc.dwFlags = 0;

	// Unnecessary if a DSound driver is present, but more accurate on emulated devices. Whatever.
	BufferDesc.dwFlags |= DSBCAPS_GETCURRENTPOSITION2;

	//BufferDesc.dwFlags |= DSBCAPS_CTRLPAN;				// Might not want this for music, UI sounds, etc. (in fact, probably only want for strictly specified in-game positional audio)
	//BufferDesc.dwFlags |= DSBCAPS_CTRLVOLUME;
	//BufferDesc.dwFlags |= DSBCAPS_CTRLFREQUENCY;		// Probably will never want this

	// For notifications.
	BufferDesc.dwFlags |= DSBCAPS_CTRLPOSITIONNOTIFY;

	BufferDesc.dwBufferBytes = Header.Subchunk2Size;

	BufferDesc.lpwfxFormat = &Format;

	IDirectSoundBuffer* NewAbstractBuffer = NULL;
	Buffer = NULL;

	Audio::mDevice->CreateSoundBuffer(&BufferDesc, &NewAbstractBuffer, NULL);
	NewAbstractBuffer->QueryInterface(IID_IDirectSoundBuffer8, (void**)&Buffer);
	NewAbstractBuffer->Release();




	void* voidPtr1 = NULL;
	DWORD NumBytes1 = 0;
	void* voidPtr2 = NULL;
	DWORD NumBytes2 = 0;

	BufferCaps.dwSize = sizeof(BufferCaps);
	Buffer->GetCaps(&BufferCaps);
	Buffer->Lock(
		0,
		0,
		&voidPtr1,
		&NumBytes1,
		&voidPtr2,
		&NumBytes2,
		DSBLOCK_ENTIREBUFFER
		);



	// Can I get away with a naive copy here?
	fread_s(voidPtr1, NumBytes1, sizeof(byte), BufferDesc.dwBufferBytes, InFile);



	Buffer->Unlock(voidPtr1, NumBytes1, voidPtr2, NumBytes2);
	
	Buffer->SetCurrentPosition(0);
}

SoundInst* Sound::Play()
{
	SoundInst* RetVal = new SoundInst();
	RetVal->Play(this);
	return RetVal;
}



SoundInst::SoundInst()
{
	DupBuffer = NULL;
	hMyEvent = NULL;
}

SoundInst::~SoundInst()
{
	SafeRelease(DupBuffer);
}

void SoundInst::Play(Sound* Template)
{
	IDirectSoundBuffer* NewAbstractBuffer = NULL;
	DupBuffer = NULL;

	HRESULT Res1 = Audio::mDevice->DuplicateSoundBuffer(Template->Buffer, &NewAbstractBuffer);
	HRESULT Res2 = NewAbstractBuffer->QueryInterface(IID_IDirectSoundBuffer8, (void**)&DupBuffer);
	NewAbstractBuffer->Release();

	



	static int SoundInstNum = 0;
	static char EventName[64];
	sprintf_s(EventName, 64, "SOUND_INST_%08x", SoundInstNum);
	hMyEvent = CreateEvent(NULL, false, false, EventName);


	LPDIRECTSOUNDNOTIFY8 lpDsNotify; 
	DSBPOSITIONNOTIFY PositionNotify;

	HRESULT Res4 = DupBuffer->QueryInterface(IID_IDirectSoundNotify8, (LPVOID*)&lpDsNotify);
	Res4;

	PositionNotify.dwOffset = DSBPN_OFFSETSTOP;
	PositionNotify.hEventNotify = hMyEvent;
	lpDsNotify->SetNotificationPositions(1, &PositionNotify);
	lpDsNotify->Release();

	// How do I test this event? I guess I need a separate thread for that? Or for each sound instance?

	HANDLE NewThreadHandle = CreateThread(
		NULL,//lpThreadAttributes,
		0,//dwStackSize,
		OnFinished,
		(void*)this,
		0,//dwCreationFlags,
		NULL);





	HRESULT Res3 = DupBuffer->Play(0, 0, 0);
}

DWORD WINAPI SoundInst::OnFinished(void* data)
{
	SoundInst* Context = (SoundInst*)(data);

	WaitForSingleObject(Context->hMyEvent, INFINITE);

	SafeDelete(Context);

	return 0;
}
