#include "NMat4.h"

#include "NMat4.h"
#include <math.h>

NMat4::NMat4()
{
	// Return the identity matrix
	val[0] = 1.0f;
	val[1] = 0.0f;
	val[2] = 0.0f;
	val[3] = 0.0f;
	val[4] = 0.0f;
	val[5] = 1.0f;
	val[6] = 0.0f;
	val[7] = 0.0f;
	val[8] = 0.0f;
	val[9] = 0.0f;
	val[10] = 1.0f;
	val[11] = 0.0f;
	val[12] = 0.0f;
	val[13] = 0.0f;
	val[14] = 0.0f;
	val[15] = 1.0f;
}

NMat4::NMat4(float inval[16])
{
	// return a matrix with the passed-in values
	val[0] = inval[0];
	val[1] = inval[1];
	val[2] = inval[2];
	val[3] = inval[3];
	val[4] = inval[4];
	val[5] = inval[5];
	val[6] = inval[6];
	val[7] = inval[7];
	val[8] = inval[8];
	val[9] = inval[9];
	val[10] = inval[10];
	val[11] = inval[11];
	val[12] = inval[12];
	val[13] = inval[13];
	val[14] = inval[14];
	val[15] = inval[15];
}

NMat4 NMat4::operator +(const NMat4& m) const
{
	// return the sum of this matrix and the given matrix
	float retVal[16] = {val[0] + m.val[0],
						val[1] + m.val[1],
						val[2] + m.val[2],
						val[3] + m.val[3],
						val[4] + m.val[4],
						val[5] + m.val[5],
						val[6] + m.val[6],
						val[7] + m.val[7],
						val[8] + m.val[8],
						val[9] + m.val[9],
						val[10] + m.val[10],
						val[11] + m.val[11],
						val[12] + m.val[12],
						val[13] + m.val[13],
						val[14] + m.val[14],
						val[15] + m.val[15]};

	NMat4 result(retVal);
	return result;
}

NMat4 NMat4::operator -(const NMat4& m) const
{
	// return the difference of this matrix and the given matrix
	float retVal[16] = {val[0] - m.val[0],
						val[1] - m.val[1],
						val[2] - m.val[2],
						val[3] - m.val[3],
						val[4] - m.val[4],
						val[5] - m.val[5],
						val[6] - m.val[6],
						val[7] - m.val[7],
						val[8] - m.val[8],
						val[9] - m.val[9],
						val[10] - m.val[10],
						val[11] - m.val[11],
						val[12] - m.val[12],
						val[13] - m.val[13],
						val[14] - m.val[14],
						val[15] - m.val[15]};

	NMat4 result(retVal);
	return result;
}

NMat4 NMat4::operator *(const NMat4& m) const
{
	// return the product of this matrix and the given matrix
	float retVal[16] = {val[0] * m.val[0] + val[1] * m.val[4] +
						val[2] * m.val[8] + val[3] * m.val[12],
						val[0] * m.val[1] + val[1] * m.val[5] +
						val[2] * m.val[9] + val[3] * m.val[13],
						val[0] * m.val[2] + val[1] * m.val[6] +
						val[2] * m.val[10] + val[3] * m.val[14],
						val[0] * m.val[3] + val[1] * m.val[7] +
						val[2] * m.val[11] + val[3] * m.val[15],
						val[4] * m.val[0] + val[5] * m.val[4] +
						val[6] * m.val[8] + val[7] * m.val[12],
						val[4] * m.val[1] + val[5] * m.val[5] +
						val[6] * m.val[9] + val[7] * m.val[13],
						val[4] * m.val[2] + val[5] * m.val[6] +
						val[6] * m.val[10] + val[7] * m.val[14],
						val[4] * m.val[3] + val[5] * m.val[7] +
						val[6] * m.val[11] + val[7] * m.val[15],
						val[8] * m.val[0] + val[9] * m.val[4] +
						val[10] * m.val[8] + val[11] * m.val[12],
						val[8] * m.val[1] + val[9] * m.val[5] +
						val[10] * m.val[9] + val[11] * m.val[13],
						val[8] * m.val[2] + val[9] * m.val[6] +
						val[10] * m.val[10] + val[11] * m.val[14],
						val[8] * m.val[3] + val[9] * m.val[7] +
						val[10] * m.val[11] + val[11] * m.val[15],
						val[12] * m.val[0] + val[13] * m.val[4] +
						val[14] * m.val[8] + val[15] * m.val[12],
						val[12] * m.val[1] + val[13] * m.val[5] +
						val[14] * m.val[9] + val[15] * m.val[13],
						val[12] * m.val[2] + val[13] * m.val[6] +
						val[14] * m.val[10] + val[15] * m.val[14],
						val[12] * m.val[3] + val[13] * m.val[7] +
						val[14] * m.val[11] + val[15] * m.val[15]};

	NMat4 result(retVal);
	return result;
}

NVec2 NMat4::operator *(const NVec2& v) const
{
	// return the product of this matrix and the given vector
	return NVec2((val[0] * v.x + val[1] * v.y + val[2] + val[3]),	// Not sure this is correct; for NVec3s, we assume
				(val[4] * v.x + val[5] * v.y + val[6] + val[7]));	// w is 1.0f, but what about z for NVec2s?
}

NVec3 NMat4::operator *(const NVec3& v) const
{
	// return the product of this matrix and the given vector
	return NVec3((val[0] * v.x + val[1] * v.y + val[2] * v.z + val[3]),
				(val[4] * v.x + val[5] * v.y + val[6] * v.z + val[7]),
				(val[8] * v.x + val[9] * v.y + val[10] * v.z + val[11]));
}

NVec4 NMat4::operator *(const NVec4& v) const
{
	// return the product of this matrix and the given vector
	return NVec4((val[0] * v.x + val[1] * v.y + val[2] * v.z + val[3] * v.w),
				(val[4] * v.x + val[5] * v.y + val[6] * v.z + val[7] * v.w),
				(val[8] * v.x + val[9] * v.y + val[10] * v.z + val[11] * v.w),
				(val[12] * v.x + val[13] * v.y + val[14] * v.z + val[15] * v.w));
}

NMat4 NMat4::operator *(float s) const
{
	// return the product of this matrix and the given scalar
	float retVal[16] = {val[0] * s,
						val[1] * s,
						val[2] * s,
						val[3] * s,
						val[4] * s,
						val[5] * s,
						val[6] * s,
						val[7] * s,
						val[8] * s,
						val[9] * s,
						val[10] * s,
						val[11] * s,
						val[12] * s,
						val[13] * s,
						val[14] * s,
						val[15] * s};

	NMat4 result(retVal);
	return result;
}

NMat4 NMat4::operator /(float s) const
{
	float is = 1.0f / s;
	// return the quotient of this matrix and the given scalar
	float retVal[16] = {val[0] * is,
						val[1] * is,
						val[2] * is,
						val[3] * is,
						val[4] * is,
						val[5] * is,
						val[6] * is,
						val[7] * is,
						val[8] * is,
						val[9] * is,
						val[10] * is,
						val[11] * is,
						val[12] * is,
						val[13] * is,
						val[14] * is,
						val[15] * is};

	NMat4 result(retVal);
	return result;
}

void NMat4::operator +=(const NMat4& m)
{
	// add the given matrix to this one
	val[0] += m.val[0];
	val[1] += m.val[1];
	val[2] += m.val[2];
	val[3] += m.val[3];
	val[4] += m.val[4];
	val[5] += m.val[5];
	val[6] += m.val[6];
	val[7] += m.val[7];
	val[8] += m.val[8];
	val[9] += m.val[9];
	val[10] += m.val[10];
	val[11] += m.val[11];
	val[12] += m.val[12];
	val[13] += m.val[13];
	val[14] += m.val[14];
	val[15] += m.val[15];
}

void NMat4::operator -=(const NMat4& m)
{
	// subtract the given matrix from this one
	val[0] -= m.val[0];
	val[1] -= m.val[1];
	val[2] -= m.val[2];
	val[3] -= m.val[3];
	val[4] -= m.val[4];
	val[5] -= m.val[5];
	val[6] -= m.val[6];
	val[7] -= m.val[7];
	val[8] -= m.val[8];
	val[9] -= m.val[9];
	val[10] -= m.val[10];
	val[11] -= m.val[11];
	val[12] -= m.val[12];
	val[13] -= m.val[13];
	val[14] -= m.val[14];
	val[15] -= m.val[15];
}

void NMat4::operator *=(const NMat4& m)
{

	float newval[16];

	newval[0] = val[0] * m.val[0] + val[1] * m.val[4] +
				val[2] * m.val[8] + val[3] * m.val[12];
	newval[1] = val[0] * m.val[1] + val[1] * m.val[5] +
				val[2] * m.val[9] + val[3] * m.val[13];
	newval[2] = val[0] * m.val[2] + val[1] * m.val[6] +
				val[2] * m.val[10] + val[3] * m.val[14];
	newval[3] = val[0] * m.val[3] + val[1] * m.val[7] +
				val[2] * m.val[11] + val[3] * m.val[15];
	newval[4] = val[4] * m.val[0] + val[5] * m.val[4] +
				val[6] * m.val[8] + val[7] * m.val[12];
	newval[5] = val[4] * m.val[1] + val[5] * m.val[5] +
				val[6] * m.val[9] + val[7] * m.val[13];
	newval[6] = val[4] * m.val[2] + val[5] * m.val[6] +
				val[6] * m.val[10] + val[7] * m.val[14];
	newval[7] = val[4] * m.val[3] + val[5] * m.val[7] +
				val[6] * m.val[11] + val[7] * m.val[15];
	newval[8] = val[8] * m.val[0] + val[9] * m.val[4] +
				val[10] * m.val[8] + val[11] * m.val[12];
	newval[9] = val[8] * m.val[1] + val[9] * m.val[5] +
				val[10] * m.val[9] + val[11] * m.val[13];
	newval[10] = val[8] * m.val[2] + val[9] * m.val[6] +
				 val[10] * m.val[10] + val[11] * m.val[14];
	newval[11] = val[8] * m.val[3] + val[9] * m.val[7] +
				 val[10] * m.val[11] + val[11] * m.val[15];
	newval[12] = val[12] * m.val[0] + val[13] * m.val[4] +
				 val[14] * m.val[8] + val[15] * m.val[12];
	newval[13] = val[12] * m.val[1] + val[13] * m.val[5] +
				 val[14] * m.val[9] + val[15] * m.val[13];
	newval[14] = val[12] * m.val[2] + val[13] * m.val[6] +
				 val[14] * m.val[10] + val[15] * m.val[14];
	newval[15] = val[12] * m.val[3] + val[13] * m.val[7] +
				 val[14] * m.val[11] + val[15] * m.val[15];

	val[0] = newval[0];
	val[1] = newval[1];
	val[2] = newval[2];
	val[3] = newval[3];
	val[4] = newval[4];
	val[5] = newval[5];
	val[6] = newval[6];
	val[7] = newval[7];
	val[8] = newval[8];
	val[9] = newval[9];
	val[10] = newval[10];
	val[11] = newval[11];
	val[12] = newval[12];
	val[13] = newval[13];
	val[14] = newval[14];
	val[15] = newval[15];
}

void NMat4::operator *=(float s)
{
	// multiply this matrix by the given scalar
	val[0] *= s;
	val[1] *= s;
	val[2] *= s;
	val[3] *= s;
	val[4] *= s;
	val[5] *= s;
	val[6] *= s;
	val[7] *= s;
	val[8] *= s;
	val[9] *= s;
	val[10] *= s;
	val[11] *= s;
	val[12] *= s;
	val[13] *= s;
	val[14] *= s;
	val[15] *= s;
}

void NMat4::operator /=(float s)
{
	float is = 1.0f / s;

	// divide this matrix by the given scalar
	val[0] *= is;
	val[1] *= is;
	val[2] *= is;
	val[3] *= is;
	val[4] *= is;
	val[5] *= is;
	val[6] *= is;
	val[7] *= is;
	val[8] *= is;
	val[9] *= is;
	val[10] *= is;
	val[11] *= is;
	val[12] *= is;
	val[13] *= is;
	val[14] *= is;
	val[15] *= is;
}

NMat4 NMat4::getTranspose() const
{
	// return the transpose of this matrix
	float retVal[16] = {val[0],
						val[4],
						val[8],
						val[12],
						val[1],
						val[5],
						val[9],
						val[13],
						val[2],
						val[6],
						val[10],
						val[14],
						val[3],
						val[7],
						val[11],
						val[15]};

	NMat4 result(retVal);
	return result;
}

float NMat4::cofactor(int i, int j) const
{
	// return the specified cofactor of this matrix
	int r0, r1, r2, c0, c1, c2;
	r0 = (i < 1) ? 4 : 0;
	r1 = (i < 2) ? 8 : 4;
	r2 = (i < 3) ? 12 : 8;
	c0 = (j < 1) ? 1 : 0;
	c1 = (j < 2) ? 2 : 1;
	c2 = (j < 3) ? 3 : 2;

	float sign = pow(-1.0f, float(i+j));
	float term1 = (val[r1+c1] * val[r2+c2]) - (val[r1+c2] * val[r2+c1]);
	float term2 = (val[r1+c0] * val[r2+c2]) - (val[r1+c2] * val[r2+c0]);
	float term3 = (val[r1+c0] * val[r2+c1]) - (val[r1+c1] * val[r2+c0]);
	return sign * ((val[r0+c0] * term1) - (val[r0+c1] * term2) + (val[r0+c2] * term3));
}

NMat4 NMat4::getInverse() const
{
	// return the inverse of this matrix

	float cf[16];

	cf[0] = cofactor(0, 0);
	cf[1] = cofactor(0, 1);
	cf[2] = cofactor(0, 2);
	cf[3] = cofactor(0, 3);
	cf[4] = cofactor(1, 0);
	cf[5] = cofactor(1, 1);
	cf[6] = cofactor(1, 2);
	cf[7] = cofactor(1, 3);
	cf[8] = cofactor(2, 0);
	cf[9] = cofactor(2, 1);
	cf[10] = cofactor(2, 2);
	cf[11] = cofactor(2, 3);
	cf[12] = cofactor(3, 0);
	cf[13] = cofactor(3, 1);
	cf[14] = cofactor(3, 2);
	cf[15] = cofactor(3, 3);

	NMat4 cofactor(cf);

	NMat4 adj = cofactor.getTranspose();

	float det = getDeterminant();

	NMat4 inv = adj / det;

	return inv;
}

float NMat4::getDeterminant() const
{
	// return the determinant of this matrix

	// pre-calculate some values to make this faster
	float a = (val[10] * val[15]) - (val[11] * val[14]);
	float b = (val[9] * val[15]) - (val[11] * val[13]);
	float c = (val[9] * val[14]) - (val[10] * val[13]);
	float d = (val[8] * val[15]) - (val[11] * val[12]);
	float e = (val[8] * val[14]) - (val[10] * val[12]);
	float f = (val[8] * val[13]) - (val[9] * val[12]);

	float det = (val[0] * ((val[5] * a) - (val[6] * b) + (val[7] * c))) -
				(val[1] * ((val[4] * a) - (val[6] * d) + (val[7] * e))) +
				(val[2] * ((val[4] * b) - (val[5] * d) + (val[7] * f))) -
				(val[3] * ((val[4] * c) - (val[5] * e) + (val[6] * f)));

	return det;
}

// The following functions generate transform matrices

NMat4 CreateScaleMatrix(const NVec3& s)
{
	float transformValues[16] = {s.x,0.0f,0.0f,0.0f,
								   0.0f,s.y,0.0f,0.0f,
								   0.0f,0.0f,s.z,0.0f,
								   0.0f,0.0f,0.0f,1.0f};

	return NMat4(transformValues);
}

NMat4 CreateScaleMatrix(float s)
{
	float transformValues[16] = {s,0.0f,0.0f,0.0f,
								   0.0f,s,0.0f,0.0f,
								   0.0f,0.0f,s,0.0f,
								   0.0f,0.0f,0.0f,1.0f};

	return NMat4(transformValues);
}

NMat4 CreateRotationMatrixX(float theta)
{
	float sn = sin(theta);
	float cs = cos(theta);
	float transformValues[16] = {1.0f,0.0f,0.0f,0.0f,
								   0.0f,cs,-sn,0.0f,
								   0.0f,sn,cs,0.0f,
								   0.0f,0.0f,0.0f,1.0f};

	return NMat4(transformValues);
}

NMat4 CreateRotationMatrixY(float theta)
{
	float sn = sin(theta);
	float cs = cos(theta);
	float transformValues[16] = {cs,0.0f,sn,0.0f,
								   0.0f,1.0f,0.0f,0.0f,
								   -sn,0.0f,cs,0.0f,
								   0.0f,0.0f,0.0f,1.0f};

	return NMat4(transformValues);
}

NMat4 CreateRotationMatrixZ(float theta)
{
	float sn = sin(theta);
	float cs = cos(theta);
	float transformValues[16] = {cs,-sn,0.0f,0.0f,
								   sn,cs,0.0f,0.0f,
								   0.0f,0.0f,1.0f,0.0f,
								   0.0f,0.0f,0.0f,1.0f};

	return NMat4(transformValues);
}

NMat4 CreateRotationMatrix(const NVec3& N, float theta)
{
	NVec3 dir = Normalize(N);
	float cos_t = cos(theta);
	float t = 1.0f - cos_t;
	float sin_t = sin(theta);
	float transformValues[16] = {(dir.x * dir.x * t) + cos_t,
									(dir.x * dir.y * t) - (dir.z * sin_t),
									(dir.x * dir.z * t) + (dir.y * sin_t),
									0.0f,
								   (dir.x * dir.y * t) + (dir.z * sin_t),
									(dir.y * dir.y * t) + cos_t,
									(dir.y * dir.z * t) - (dir.x * sin_t),
									0.0f,
								   (dir.x * dir.z * t) - (dir.y * sin_t),
									(dir.y * dir.z * t) + (dir.x * sin_t),
									(dir.z * dir.z * t) + cos_t,
									0.0f,
								   0.0f,0.0f,0.0f,1.0f};

	return NMat4(transformValues);
}

NMat4 CreateRotationMatrix(const NVec3& x, const NVec3& y, const NVec3& z)
{
	// TODO: Assert that x, y, and z are orthonormal?
	// Or enforce orthonormality?

	float transformValues[16] =
	{
		x.x,x.y,x.z,0.0f,
		y.x,y.y,y.z,0.0f,
		z.x,z.y,z.z,0.0f,
		0.0f,0.0f,0.0f,1.0f
	};

	return NMat4(transformValues);
}

void NMat4::GetAxes(NVec3& x, NVec3& y, NVec3& z)
{
	x = NVec3(val[0], val[1], val[2]);
	y = NVec3(val[4], val[5], val[6]);
	z = NVec3(val[8], val[9], val[10]);
}

// For the camera-related function, see CreateProjMatrix.
NMat4 CreateProjectionMatrix(const NVec3& n)
{
	float transformValues[16] = {1 - (n.x * n.x), -(n.x * n.y), -(n.x * n.z), 0.0f,
								   -(n.x * n.y), 1 - (n.y * n.y), -(n.y * n.z), 0.0f,
								   -(n.x * n.z), -(n.y * n.z), 1 - (n.z * n.z), 0.0f,
								   0.0f,0.0f,0.0f,1.0f};

	return NMat4(transformValues);
}

NMat4 CreateReflectionMatrix(const NVec3& n)
{
	float transformValues[16] = {1.0f - 2.0f*(n.x * n.x), -2.0f*(n.x * n.y), -2.0f*(n.x * n.z), 0.0f,
								   -2.0f*(n.x * n.y), 1.0f - 2.0f*(n.y * n.y), -2.0f*(n.y * n.z), 0.0f,
								   -2.0f*(n.x * n.z), -2.0f*(n.y * n.z), 1.0f - 2.0f*(n.z * n.z), 0.0f,
								   0.0f,0.0f,0.0f,1.0f};

	return NMat4(transformValues);
}

NMat4 CreateTranslationMatrix(const NVec3& d)
{
	float transformValues[16] = {1.0f,0.0f,0.0f,0.0f,
								   0.0f,1.0f,0.0f,0.0f,
								   0.0f,0.0f,1.0f,0.0f,
								   d.x,d.y,d.z,1.0f};

	return NMat4(transformValues);
}

NMat4 CreateProjMatrix(float fov, float aspectRatio, float nearClip, float farClip, float targetAspectRatio)
{
	// For future reference, this matches the standard D3D Matrix.PerspectiveFovRH() exactly.

	float yScale = 1.0f;
	float xScale = 1.0f;

	// 04/27/2011: Playing around with the concept of using a horizontal FOV for tallscreen modes.
	// 01/17/2011: Provide a target aspect ratio about which the projection matrix will be sized.
	if (aspectRatio >= targetAspectRatio)
	{
		// Widescreen
		yScale = 1.0f / tan(fov / 2.0f);
		xScale = yScale / aspectRatio;
	}
	else
	{
		// Tallscreen
		float h_fov = 2.0f * atan(targetAspectRatio * tan(fov / 2.0f));

		xScale = 1.0f / tan(h_fov / 2.0f);
		yScale = xScale * aspectRatio;
	}

	float z1 = farClip / (nearClip - farClip);
	float z2 = nearClip * farClip / (nearClip - farClip);

	float transformValues[16] = {xScale,	0.0f,	0.0f,	0.0f,
								   0.0f,	yScale,	0.0f,	0.0f,
								   0.0f,	0.0f,	z1,		-1.0f,
								   0.0f,	0.0f,	z2,		0.0f};

	return NMat4(transformValues);
}

NMat4 CreateViewMatrix(const NVec3& pos, const NVec3& forward, const NVec3& left, const NVec3& up)
{
	float transformValues[16] = {-left.x,			up.x,			-forward.x,			0.0f,
								   -left.y,			up.y,			-forward.y,			0.0f,
								   -left.z,			up.z,			-forward.z,			0.0f,
								   Dot(left, pos),	-Dot(up, pos),	Dot(forward, pos),	1.0f};


	return NMat4(transformValues);
}

NMat4 CreateOrthoMatrix(float left, float right, float top, float bottom, float nearClip, float farClip)
{
	// TODO: Add targetAspectRatio here, too?

	float z1 = 1.0f / (nearClip - farClip);
	float z2 = nearClip * z1;

	float transformValues[16] = {2.0f / (right - left),	0.0f,	0.0f,	0.0f,
								   0.0f,	2.0f / (top - bottom),	0.0f,	0.0f,
								   0.0f,	0.0f,	z1,	0.0f,
								   (left + right) / (left - right),	(top + bottom) / (bottom - top),	z2,	1.0f};

	return NMat4(transformValues);
}
