#include <memory.h>
#include "NMathFtns.h"

void DCT(float* complexVals, int numVals)
{
	float* dup = new float[numVals];
	memcpy_s(dup, sizeof(float) * numVals, complexVals, sizeof(float) * numVals);

	float RecNumVals = 1.0f / numVals;
	float SqrtRecNumVals = sqrt(RecNumVals);
	float Sqrt2RecNumVals = sqrt(2.0f * RecNumVals);

	for (int i = 0; i < numVals; ++i)
	{
		float sum = 0.0f;
		//NComplex sum(0.0f, 0.0f);
		for (int j = 0; j < numVals; ++j)
		{
			sum += (dup[j] * cos(0.5f * PI * RecNumVals * float(i) * (2.0f * float(j) + 1.0f)));

			//float x = -(float(i) * float(j) * TwoPiOverNumVals);
			//float sn = sin(x);
			//float cs = cos(x);
			//sum += (dup[j] * NComplex(cs, sn));
		}
		if (i == 0)
		{
			complexVals[i] = sum * SqrtRecNumVals;
		}
		else
		{
			complexVals[i] = sum * Sqrt2RecNumVals;
		}
	}

	delete[] dup;
}

void DCTi(float* complexVals, int numVals)
{
	float* dup = new float[numVals];
	memcpy_s(dup, sizeof(float) * numVals, complexVals, sizeof(float) * numVals);

	float RecNumVals = 1.0f / numVals;
	float SqrtRecNumVals = sqrt(RecNumVals);
	float Sqrt2RecNumVals = sqrt(2.0f * RecNumVals);

	for (int j = 0; j < numVals; ++j)
	{
		float sum = 0.0f;
		//NComplex sum(0.0f, 0.0f);
		for (int i = 0; i < numVals; ++i)
		{
			float foo = (dup[i] * cos(0.5f * PI * RecNumVals * float(i) * (2.0f * float(j) + 1.0f)));

			if (i == 0)
			{
				foo *= SqrtRecNumVals;
			}
			else
			{
				foo *= Sqrt2RecNumVals;
			}

			sum += foo;

			//float x = (float(i) * float(j) * TwoPiOverNumVals);
			//float sn = sin(x);
			//float cs = cos(x);
			//sum += (dup[j] * NComplex(cs, sn));
		}
		complexVals[j] = sum;
	}

	delete[] dup;
}

void DCT2(float** complexVals, int numValsX, int numValsY)
{
	float** dup = new float*[numValsX];
	for (int i = 0; i < numValsX; ++i)
	{
		dup[i] = new float[numValsY];
		memcpy_s(dup[i], sizeof(float) * numValsY, complexVals[i], sizeof(float) * numValsY);
	}

	float RecNumValsX = 1.0f / (numValsX);
	float RecNumValsY = 1.0f / (numValsY);
	float SqrtRecNumValsX = sqrt(RecNumValsX);
	float SqrtRecNumValsY = sqrt(RecNumValsY);
	float Sqrt2RecNumValsX = sqrt(2.0f * RecNumValsX);
	float Sqrt2RecNumValsY = sqrt(2.0f * RecNumValsY);

	for (int i = 0; i < numValsX; ++i)
	{
		for (int j = 0; j < numValsY; ++j)
		{
			float sum = 0.0f;
			//NComplex sum(0.0f, 0.0f);
			for (int k = 0; k < numValsX; ++k)
			{
				for (int l = 0; l < numValsY; ++l)
				{
					sum += (dup[k][l] *
						cos(0.5f * PI * RecNumValsX * float(i) * (2.0f * float(k) + 1.0f)) *
						cos(0.5f * PI * RecNumValsY * float(j) * (2.0f * float(l) + 1.0f)));

					//float x = -(float(i) * float(j) * float(k) * float(l) * TwoPiOverNumVals);
					//float sn = sin(x);
					//float cs = cos(x);
					//sum += (dup[k][l] * NComplex(cs, sn));
				}
			}
			if (i == 0)
			{
				complexVals[i][j] = sum * SqrtRecNumValsX;
			}
			else
			{
				complexVals[i][j] = sum * Sqrt2RecNumValsX;
			}
			if (j == 0)
			{
				complexVals[i][j] *= SqrtRecNumValsY;
			}
			else
			{
				complexVals[i][j] *= Sqrt2RecNumValsY;
			}
		}
	}

	for (int i = 0; i < numValsX; ++i)
	{
		delete[] dup[i];
	}
	delete[] dup;
}

void DCT2i(float** complexVals, int numValsX, int numValsY)
{
	float** dup = new float*[numValsX];
	for (int i = 0; i < numValsX; ++i)
	{
		dup[i] = new float[numValsY];
		memcpy_s(dup[i], sizeof(float) * numValsY, complexVals[i], sizeof(float) * numValsY);
	}

	float RecNumValsX = 1.0f / (numValsX);
	float RecNumValsY = 1.0f / (numValsY);
	float SqrtRecNumValsX = sqrt(RecNumValsX);
	float SqrtRecNumValsY = sqrt(RecNumValsY);
	float Sqrt2RecNumValsX = sqrt(2.0f * RecNumValsX);
	float Sqrt2RecNumValsY = sqrt(2.0f * RecNumValsY);

	for (int k = 0; k < numValsX; ++k)
	{
		for (int l = 0; l < numValsY; ++l)
		{
			float sum = 0.0f;
			//NComplex sum(0.0f, 0.0f);
			for (int i = 0; i < numValsX; ++i)
			{
				for (int j = 0; j < numValsY; ++j)
				{
					float foo = (dup[i][j] *
						cos(0.5f * PI * RecNumValsX * float(i) * (2.0f * float(k) + 1.0f)) *
						cos(0.5f * PI * RecNumValsY * float(j) * (2.0f * float(l) + 1.0f)));

					if (i == 0)
					{
						foo *= SqrtRecNumValsX;
					}
					else
					{
						foo *= Sqrt2RecNumValsX;
					}
					if (j == 0)
					{
						foo *= SqrtRecNumValsY;
					}
					else
					{
						foo *= Sqrt2RecNumValsY;
					}

					sum += foo;

					//float x = (float(i) * float(j) * float(k) * float(l) * TwoPiOverNumVals);
					//float sn = sin(x);
					//float cs = cos(x);
					//sum += (dup[k][l] * NComplex(cs, sn));
				}
			}
			complexVals[k][l] = sum;
		}
	}

	for (int i = 0; i < numValsX; ++i)
	{
		delete[] dup[i];
	}
	delete[] dup;
}

// Operates on degrees because I'm using it for camera initialization at the moment...hurr durr?
float HorizontalFOV(float VerticalFOV, float AspectRatio)
{
	float FOVInRads = DEG2RAD * VerticalFOV;
	float HalfFOV = FOVInRads * 0.5f;
	float Tangent = tan(HalfFOV);
	float Scaled = Tangent * AspectRatio;
	float InvTan = atan(Scaled);
	float FOV = InvTan * 2.0f;
	float FOVInDegs = RAD2DEG * FOV;

	return FOVInDegs;
}

// Operates on degrees because I'm using it for camera initialization at the moment...hurr durr?
float VerticalFOV(float HorizontalFOV, float AspectRatio)
{
	float FOVInRads = DEG2RAD * HorizontalFOV;
	float HalfFOV = FOVInRads * 0.5f;
	float Tangent = tan(HalfFOV);
	float Scaled = Tangent / AspectRatio;
	float InvTan = atan(Scaled);
	float FOV = InvTan * 2.0f;
	float FOVInDegs = RAD2DEG * FOV;

	return FOVInDegs;
}

unsigned int atox(const char* InText)
{
	unsigned int RetVal = 0;

	const char* Itr = InText;

	while (Itr && *Itr)
	{
		RetVal *= 16;

		switch (*Itr)
		{
		case '0':
			RetVal += 0;
			break;

		case '1':
			RetVal += 1;
			break;

		case '2':
			RetVal += 2;
			break;

		case '3':
			RetVal += 3;
			break;

		case '4':
			RetVal += 4;
			break;

		case '5':
			RetVal += 5;
			break;

		case '6':
			RetVal += 6;
			break;

		case '7':
			RetVal += 7;
			break;

		case '8':
			RetVal += 8;
			break;

		case '9':
			RetVal += 9;
			break;

		case 'a':
		case 'A':
			RetVal += 0x0a;
			break;

		case 'b':
		case 'B':
			RetVal += 0x0b;
			break;

		case 'c':
		case 'C':
			RetVal += 0x0c;
			break;

		case 'd':
		case 'D':
			RetVal += 0x0d;
			break;

		case 'e':
		case 'E':
			RetVal += 0x0e;
			break;

		case 'f':
		case 'F':
			RetVal += 0x0f;
			break;

		default:
			break;
		}

		Itr++;
	}

	return RetVal;
}
