#include <assert.h>
#include "NMathFtns.h"
#include "NString.h"
#include "NText.h"
#include "NUtilDefs.h"

char* NString::text_null = "";

NString::NString()
{
	init_empty();
}

void NString::init_empty()
{
	// Does not call free() because this should only ever be used on construction.

	text = text_null;
	length = 0;
	prcvln = 0;
	utf8 = EU8_Unknown;
}

NString::NString(const NString& string)
{
	text = NULL;
	length = string.Copy(text);
	prcvln = CalcPercLen();
	utf8 = string.utf8;
}

NString::NString(const NHashedString& string)
{
	text = StringCopy(string.getString());
	length = (int)(strlen(string.getString()));
	prcvln = CalcPercLen();
	utf8 = EU8_Unknown;
}

NString::NString(char* string, int ExactLength)
{
	if (string)
	{
		text = StringCopy(string, ExactLength);
		length = (ExactLength < 0) ? ((int)(strlen(string))) : (Min(ExactLength, (int)strlen(string)));
		prcvln = CalcPercLen(ExactLength);
		utf8 = EU8_Unknown;
	}
	else
	{
		init_empty();
	}
}

NString::NString(const char* string, int ExactLength)
{
	if (string)
	{
		text = StringCopy(string, ExactLength);
		length = (ExactLength < 0) ? ((int)(strlen(string))) : (Min(ExactLength, (int)strlen(string)));
		prcvln = CalcPercLen(ExactLength);
		utf8 = EU8_Unknown;
	}
	else
	{
		init_empty();
	}
}

NString NString::operator =(const NString& string)
{
	free();
	length = string.Copy(text);
	prcvln = CalcPercLen();
	utf8 = string.utf8;
	return *this;
}

NString NString::operator =(const NHashedString& string)
{
	free();
	text = StringCopy(string.getString());
	length = (int)(strlen(string.getString()));
	prcvln = CalcPercLen();
	utf8 = EU8_Unknown;
	return *this;
}

NString NString::operator =(char* string)
{
	free();
	if (string)
	{
		text = StringCopy(string);
		length = (int)(strlen(string));
		prcvln = CalcPercLen();
		utf8 = EU8_Unknown;
	}
	else
	{
		init_empty();
	}
	return *this;
}

NString NString::operator =(const char* string)
{
	free();
	if (string)
	{
		text = StringCopy(string);
		length = (int)(strlen(string));
		prcvln = CalcPercLen();
		utf8 = EU8_Unknown;
	}
	else
	{
		init_empty();
	}
	return *this;
}

NString NString::operator +=(const NString& string)
{
	char* oldText = text;
	text = StringCat(**this, Length(), *string, string.Length());
	length = Length() + string.Length();
	prcvln = CalcPercLen();
	SafeStringDelete(oldText);
	return *this;
}

NString NString::operator +=(const NHashedString& string)
{
	char* oldText = text;
	text = StringCat(**this, Length(), string.getString(), (int)strlen(string.getString()));
	length = Length() + (int)strlen(string.getString());
	prcvln = CalcPercLen();
	SafeStringDelete(oldText);
	return *this;
}

NString NString::operator +=(char* string)
{
	if (string)
	{
		char* oldText = text;
		text = StringCat(**this, Length(), string, (int)strlen(string));
		length = Length() + (int)strlen(string);
		prcvln = CalcPercLen();
		SafeStringDelete(oldText);
	}
	return *this;
}

NString NString::operator +=(const char* string)
{
	if (string)
	{
		char* oldText = text;
		text = StringCat(**this, Length(), string, (int)strlen(string));
		length = Length() + (int)strlen(string);
		prcvln = CalcPercLen();
		SafeStringDelete(oldText);
	}
	return *this;
}

NString NString::operator +(const NString& string) const
{
	char* newText = StringCat(**this, Length(), *string, string.Length());
	NString RetVal(newText);
	SafeStringDelete(newText);
	return RetVal;
}

NString NString::operator +(const NHashedString& string) const
{
	char* newText = StringCat(**this, Length(), string.getString(), (int)strlen(string.getString()));
	NString RetVal(newText);
	SafeStringDelete(newText);
	return RetVal;
}

NString NString::operator +(char* string) const
{
	if (string)
	{
		char* newText = StringCat(**this, Length(), string, (int)strlen(string));
		NString RetVal(newText);
		SafeStringDelete(newText);
		return RetVal;
	}
	else
	{
		return *this;
	}
}

NString NString::operator +(const char* string) const
{
	if (string)
	{
		char* newText = StringCat(**this, Length(), string, (int)strlen(string));
		NString RetVal(newText);
		SafeStringDelete(newText);
		return RetVal;
	}
	else
	{
		return *this;
	}
}

bool NString::operator ==(const NString& string) const
{
	return (strcmp(text, *string) == 0);
}

bool NString::operator ==(const NHashedString& string) const
{
	return (strcmp(text, string.getString()) == 0);
}

bool NString::operator ==(char* string) const
{
	if (string)
	{
		return (strcmp(text, string) == 0);
	}
	else
	{
		return (length == 0);
	}
}

bool NString::operator ==(const char* string) const
{
	if (string)
	{
		return (strcmp(text, string) == 0);
	}
	else
	{
		return (length == 0);
	}
}

bool NString::operator ^(const NString& string) const
{
	return (_stricmp(text, *string) == 0);
}

bool NString::operator ^(const NHashedString& string) const
{
	return (_stricmp(text, string.getString()) == 0);
}

bool NString::operator ^(char* string) const
{
	if (string)
	{
		return (_stricmp(text, string) == 0);
	}
	else
	{
		return (length == 0);
	}
}

bool NString::operator ^(const char* string) const
{
	if (string)
	{
		return (_stricmp(text, string) == 0);
	}
	else
	{
		return (length == 0);
	}
}

bool NString::operator !=(const NString& string) const
{
	return (strcmp(text, *string) != 0);
}

bool NString::operator !=(const NHashedString& string) const
{
	return (strcmp(text, string.getString()) != 0);
}

bool NString::operator !=(char* string) const
{
	if (string)
	{
		return (strcmp(text, string) != 0);
	}
	else
	{
		return (length != 0);
	}
}

bool NString::operator !=(const char* string) const
{
	if (string)
	{
		return (strcmp(text, string) != 0);
	}
	else
	{
		return (length != 0);
	}
}

int NString::Copy(char*& OutText) const
{
	assert(OutText == NULL);

	// Just in case, for release builds:
	SafeStringDelete(OutText);

	int Len = Length();

	if (text)
	{
		OutText = new char[Len+1];
		memcpy_s(OutText, sizeof(char) * Len, text, sizeof(char) * Len);
		OutText[Len] = 0;
	}

	return Len;
}

void NString::free()
{
	SafeStringDelete(text);
	text = NULL;
	length = -1;
	prcvln = -1;
}

NString NString::Left(int InCount) const
{
	NString RetVal;

	size_t Count = 0;
	if (InCount >= 0)
	{
		Count = size_t(InCount);
	}
	else
	{
		Count = size_t(length + InCount);
	}

	if (text)
	{
		size_t Len = (size_t)Min((int)Count, length);
		char* CharArray = new char[Len+1];
		memcpy_s(CharArray, sizeof(char) * Len, text, sizeof(char) * Len);
		CharArray[Len] = 0;
		RetVal.free();		// Guh, have to free empty string
		RetVal.text = CharArray;
		RetVal.length = (int)Len;
	}

	return RetVal;
}

NString NString::Right(int InCount) const
{
	NString RetVal;

	size_t Count = 0;
	if (InCount >= 0)
	{
		Count = size_t(InCount);
	}
	else
	{
		Count = size_t(length + InCount);
	}

	if (text)
	{
		size_t Len = (size_t)Min((int)Count, length);
		char* CharArray = new char[Len+1];
		memcpy_s(CharArray, sizeof(char) * Len, text + (length - Count), sizeof(char) * Len);
		CharArray[Len] = 0;
		RetVal.free();		// Guh, have to free empty string
		RetVal.text = CharArray;
		RetVal.length = (int)Len;
	}

	return RetVal;
}

NString NString::Mid(size_t Pos, size_t Count) const
{
	NString RetVal;

	if (text)
	{
		size_t Len = (size_t)Min((int)Count, length);
		char* CharArray = new char[Len+1];
		memcpy_s(CharArray, sizeof(char) * Len, text + size_t(Clamp((int)Pos, 0, length)), sizeof(char) * Len);
		CharArray[Len] = 0;
		RetVal.free();		// Guh, have to free empty string
		RetVal.text = CharArray;
		RetVal.length = (int)Len;
	}

	return RetVal;
}

NString NString::Trimmed() const
{
	NString RetVal;

	if (text)
	{
		char* StartChar = text;
		char* EndChar = text + length;

		while (StartChar < EndChar && IsWhitespace(*StartChar))
		{
			StartChar++;
		}

		while (EndChar > StartChar && (IsWhitespace(*(EndChar-1))))
		{
			EndChar--;
		}

		INT_PTR Len = (EndChar - StartChar);

		assert(Len >= 0);

		char* CharArray = new char[Len+1];
		memcpy_s(CharArray, sizeof(char) * Len, StartChar, sizeof(char) * Len);
		CharArray[Len] = 0;
		RetVal.free();		// Guh, have to free empty string
		RetVal.text = CharArray;
		RetVal.length = (int)Len;
	}

	return RetVal;
}

int NString::Split(const NString& Partition, std::vector<NString>& OutArray) const
{
	size_t PartitionLen = Partition.Length();

	std::vector<char*> Matches;
	char* itr = text;
	if (PartitionLen > 0)
	{
		while (*itr)
		{
			bool bMatch = true;
			char* itr_partition = *Partition;
			for (size_t i = 0; i < PartitionLen && bMatch; ++i)
			{
				bMatch = (*(itr + i) ==  *(itr_partition + i));
			}

			if (bMatch)
			{
				Matches.push_back(itr);

				itr += PartitionLen;
			}
			else
			{
				itr++;
			}
		}
		Matches.push_back(itr);
	}

	OutArray.clear();

	char* LastMarker = text;
	for (int i = 0; i < (int)Matches.size(); ++i)
	{
		//char Preserved = *Matches[i];

		// This crashes if the string is const!!
		// We really need to make a copy of it.
		//*Matches[i] = 0;
		int ExactLength = (int)(Matches[i] - LastMarker);
		OutArray.push_back(NString(LastMarker, ExactLength));
		//*Matches[i] = Preserved;

		LastMarker = (Matches[i] + PartitionLen);
	}

	return (int)Matches.size();
}

NString NString::Replace(const NString& SearchTerm, const NString& ReplacementTerm) const
{
	NString RetVal;

	std::vector<NString> OutArray;
	int Parts = Split(SearchTerm, OutArray);

	for (int i = 0; i < Parts; ++i)
	{
		RetVal += OutArray[i];
		if (i < Parts - 1)
		{
			RetVal += ReplacementTerm;
		}
	}

	return RetVal;
}

int NString::Find(const NString& Term, size_t From) const
{
	size_t TermLen = Term.Length();

	char* Match = NULL;
	char* itr = text + From;
	if (TermLen > 0)
	{
		while (*itr && !Match)
		{
			bool bMatch = true;
			char* itr_term = *Term;
			for (size_t i = 0; i < TermLen && bMatch; ++i)
			{
				bMatch = (*(itr + i) ==  *(itr_term + i));
			}

			if (bMatch)
			{
				Match = itr;
			}
			else
			{
				itr++;
			}
		}
	}

	int RetVal = -1;

	if (Match)
	{
		RetVal = int(Match - text);
	}

	return RetVal;
}

void NString::ResolveAll()
{
	// Actually resolving everything would be very expensive as noted below.
	// Let's just check for UTF-8 presence instead.

	utf8 = EU8_Unknown;

	for (char* itr = text; *itr && utf8 == EU8_Unknown; ++itr)
	{
		if (*itr & 0x80)
		{
			utf8 = EU8_Yes;
		}
	}

	if (utf8 == EU8_Unknown)
	{
		utf8 = EU8_No;
	}
}

// This is prohibitively expensive to do for every loc string in debug.
// Rewrite to make faster and fix up other escape chars?
NString NString::ResolveEscapeChars(const NString& InString)
{
	NString RetVal;

	std::vector<NString> OutStringArray;

	InString.Split("\\n", OutStringArray);

	for (int i = 0; i < (int)OutStringArray.size(); ++i)
	{
		RetVal += OutStringArray[i];
		if (i < (int)OutStringArray.size() - 1)
		{
			RetVal += "\n";
		}
	}

	return RetVal;
}

NString NString::GetStripped(NString TestChars)
{
	NString RetVal;

	const char* buffer = text;
	while (*buffer)
	{
		bool bFound = false;

		for (int i = 0; i < TestChars.Length() && !bFound; ++i)
		{
			bFound = (TestChars.text[i] == *buffer);
		}

		if (!bFound)
		{
			RetVal += Text("%c", *buffer);
		}

		buffer++;
	}

	return RetVal;
}

NString NString::GetStrippedNonalphanumeric()
{
	NString RetVal;

	const char* buffer = text;
	while (*buffer)
	{
		if (IsAlphanumeric(*buffer))
		{
			RetVal += Text("%c", *buffer);
		}
		buffer++;
	}

	return RetVal;
}

NString NString::GetUpper() const
{
	NString RetVal;

	const char* buffer = text;
	while (*buffer)
	{
		char TestChar = *buffer;
		if (IsAlphaLower(TestChar))
		{
			TestChar -= 32;
		}
		RetVal += Text("%c", TestChar);
		buffer++;
	}

	return RetVal;
}

NString NString::GetLower() const
{
	NString RetVal;

	const char* buffer = text;
	while (*buffer)
	{
		char TestChar = *buffer;
		if (IsAlphaUpper(TestChar))
		{
			TestChar += 32;
		}
		RetVal += Text("%c", TestChar);
		buffer++;
	}

	return RetVal;
}

void NString::SafeStringDelete(char* InText)
{
	if (InText != text_null)
	{
		SafeDeleteArray(InText);
	}
}

int NString::GetInt() const
{
	return atoi(*(Trimmed().GetLower()));
}

float NString::GetFloat() const
{
	return (float)atof(*(Trimmed().GetLower()));
}

bool NString::GetBool() const
{
	return StringToBool(*(Trimmed().GetLower()));
}

// Helper functions for source code parsing.

// This is stupid slow.
// I need to flag start/end points and only append contents when I reach the start of a new comment or the end of the string.
NString NString::StripBlockComments()
{
	NString RetVal;

	bool bIgnoring = false;

	char* TextStart = text;

	char* buffer = text;
	while (*buffer)
	{
		if (!bIgnoring &&
			*(buffer+0) && *(buffer+0) == '/' &&
			*(buffer+1) && *(buffer+1) == '*')
		{
			// Append contents.
			char temp = *buffer;
			*buffer = 0;
			RetVal += TextStart;
			*buffer = temp;

			// Open comment
			buffer += 2;
			bIgnoring = true;
		}
		if (bIgnoring &&
			*(buffer+0) && *(buffer+0) == '*' &&
			*(buffer+1) && *(buffer+1) == '/')
		{
			// Close comment
			buffer += 2;
			bIgnoring = false;

			TextStart = buffer;
		}
		else
		{
			//if (!bIgnoring)
			//{
			//	RetVal += Text("%c", *buffer);
			//}
			buffer++;
		}
	}

	// Append final contents.
	if (!bIgnoring)
	{
		RetVal += TextStart;
	}

	return RetVal;
}

NString NString::StripLineComments()
{
	NString RetVal;

	bool bIgnoring = false;

	char* TextStart = text;

	char* buffer = text;
	while (*buffer)
	{
		if (!bIgnoring &&
			*(buffer+0) && *(buffer+0) == '/' &&
			*(buffer+1) && *(buffer+1) == '/')
		{
			// Append contents.
			char temp = *buffer;
			*buffer = 0;
			RetVal += TextStart;
			*buffer = temp;

			// Open comment
			buffer += 2;
			bIgnoring = true;
		}
		//if (bIgnoring &&
		//	*(buffer+0) && *(buffer+0) == '\r' &&
		//	*(buffer+1) && *(buffer+1) == '\n')
		if (bIgnoring &&
			*(buffer+0) && *(buffer+0) == '\n')
		{
			// Close comment
			//buffer += 2;
			//buffer += 1;
			bIgnoring = false;

			TextStart = buffer;
		}
		else
		{
			//if (!bIgnoring)
			//{
			//	RetVal += Text("%c", *buffer);
			//}
			buffer++;
		}
	}

	// Append final contents.
	if (!bIgnoring)
	{
		RetVal += TextStart;
	}

	return RetVal;
}

int NString::CalcPercLen(int ExactLength) const
{
	int RetVal = 0;

	opt_assert(length >= 0);

	int Iter = (ExactLength >= 0) ? (Min(length, ExactLength)) : length;

	for (int i = 0; i < Iter; ++i)
	{
		if ((text[i] & 0x80) != 0)
		{
			RetVal++;
			while ((text[i] & 0x80) != 0)
			{
				opt_assert(i < Iter);
				++i;
			}

			// D'oh, gotta back this up once before starting the next iteration.
			--i;
		}
		else
		{
			RetVal++;
		}
	}

	return RetVal;
}

// Takes in a desired perceived substring length and returns the number of bytes required to draw that much text, accounting for glyphs and UTF-8 characters.
int NString::PercToRealLen(int InPercLen) const
{
	int RetVal = 0;

	opt_assert(length >= 0);

	int Iter = (InPercLen >= 0) ? (Min(prcvln, InPercLen)) : prcvln;

	for (int i = 0; i < Iter; ++i)
	{
		if ((text[RetVal] & 0x80) != 0)
		{
			RetVal++;
			while ((text[RetVal] & 0x80) != 0)
			{
				RetVal++;
			}
		}
		else
		{
			RetVal++;
		}
	}

	return RetVal;
}
