#pragma once

#include <string.h>
#include <stdio.h>
#include "NUtilDefs.h"
#include <assert.h>
#include "NWindowsVersion.h"
#include "NHashedString.h"

// Asserts that can be optionally ignored at run time
#undef opt_assert
#ifdef NDEBUG	// no debug
	#define opt_assert(x) (sizeof(x))
#else
	#define opt_assert(x)																		\
	{																							\
		static bool _OPT_OUT = false;															\
		int result = -1;																		\
		bool failcond = !(x);																	\
		if (failcond && !_OPT_OUT)																\
		{																						\
			char buffer[512];																	\
			sprintf_s(buffer, 512, "Assertion failed: %s\nFile: %s\nLine: %d\n\n"				\
				"Would you like to continue to receive this message each time this assertion "	\
				"fails?  (Pressing Cancel will assert as usual.)", #x, __FILE__, __LINE__);		\
			result = MessageBox(NULL, buffer, "Dev Assert", MB_YESNOCANCEL | MB_ICONWARNING);	\
			_OPT_OUT = (result == IDNO) ? true : (result == IDYES) ? false : _OPT_OUT;			\
			if (result == IDCANCEL)																\
			{																					\
				_wassert(_CRT_WIDE(#x), _CRT_WIDE(__FILE__), __LINE__);							\
			}																					\
		}																						\
	}
#endif



#define FOLDER_SEPARATOR "\\"
#define FOLDER_SEPARATOR_CHAR '\\'



#ifdef _DEBUG
#define UNIMPLEMENTED					\
{										\
	bool _NOT_YET_IMPLEMENTED = false;	\
	opt_assert(_NOT_YET_IMPLEMENTED);	\
}
#else
#define UNIMPLEMENTED ;
#endif

#ifdef _DEBUG
#define ANCHOR				\
{							\
	int _ANCHOR_VAR = 0;	\
	_ANCHOR_VAR++;			\
}
#else
#define ANCHOR ;
#endif

#define XOR(x,y) (((x) && (!y)) || ((!x) && (y)))

#define SafeDelete(x) {if(x){delete(x);(x)=NULL;}}
#define SafeDeleteArray(x) {if(x){delete[](x);(x)=NULL;}}
#define SafeDeleteNoNull(x) {if(x){delete(x);}}
#define SafeDeleteArrayNoNull(x) {if(x){delete[](x);}}

#define SafeRelease(x) {if(x){(x)->Release();(x)=NULL;}}

#define GetMutex(x)							\
{											\
	WaitForSingleObject((x), INFINITE);		\
}

#define MemCopy(d,s,n) memcpy_s((d),(n),(s),(n))

#define MemCopyReadAdvance(d,s,n)	\
	{								\
		MemCopy((d),(s),(n));		\
		(s) += (n);					\
	}

#define MemCopyWriteAdvance(d,s,n)	\
	{								\
		MemCopy((d),(s),(n));		\
		(d) += (n);					\
	}

// These probably don't belong in a file called "NUtilDefs.h."
// Rename this?

// Functions for char* strings
struct CharArrayEq
{
	bool operator() (const char* str1, const char* str2)
	{
		return (strcmp(str1, str2) == 0);
	}
};

struct CharArrayHash
{
	uint operator() (const char* str)
	{
		uint hash = 5381;
		char* ptr = (char*)str;
		uint c = *ptr;

		while (c)
		{
			hash = ((hash << 5) + hash) + c;
			ptr++;
			c = *ptr;
		}

		return hash;
	}
};

// Functions for hashed strings
struct StrEq
{
	inline bool operator() (NHashedString str1, NHashedString str2)
	{
		return (str1 == str2);
	}
};

struct StrHash
{
	// Not a sufficient hashing function -- the vast majority of hashed string IDs are going to end in zero (first entry in hash table),
	// so when they get modded by the size of the global data table (256), they ALL end up in the first slot.
	inline uint operator() (NHashedString str)
	{
		// Assumes that the hash bits are the low bits
		return str.getID();
	}
};
